from .WorkerThread import WorkerThread
from .ChangeNotificationMessage import ChangeNotificationMessage
import json

#Ellucian documentation for this
# - https://resources.elluciancloud.com/bundle/ethos_integration_ref_apis/page/r_message_queue_api_retrieve_msgs.html

class EthosChangeNotificationPollerThreadExceptionClass(Exception):
  pass

class EthosChangeNotificationPollerThread(WorkerThread):
  clientAPIInstance = None
  loginSession = None
  pageLimit = None
  maxRequests = None
  lastProcessedID = None

  def __init__(
    self,
    clientAPIInstance,
    loginSession,
    frequency,
    pageLimit,
    maxRequests,
    lastProcessedID
  ):
    super().__init__(sleepTime=0.1, frequency=frequency)
    self.clientAPIInstance=clientAPIInstance
    self.loginSession=loginSession
    self.pageLimit=pageLimit
    self.maxRequests=maxRequests
    self.lastProcessedID = lastProcessedID

  def worker(self):
    fetchRunning = True
    numRequestsSent = 0
    while fetchRunning:
      numRequestsSent += 1
      remainingMessages = self.requestBatchOfPagesAndReturnRemainingCount()
      if remainingMessages == 0:
        fetchRunning = False
      if numRequestsSent == self.maxRequests:
        fetchRunning = False

  def requestBatchOfPagesAndReturnRemainingCount(self):
    params = {
      "limit": str(self.pageLimit)
    }
    if self.lastProcessedID is not None:
      params["lastProcessedID"] = self.lastProcessedID
    result = self.clientAPIInstance.sendGetRequest(
      url="/consume",
      params=params,
      loginSession=self.loginSession,
      injectHeadersFn=None
    )
    if result.status_code != 200:
      self.clientAPIInstance.raiseResponseException(result)

    remainingMessages = int(result.headers["x-remaining"])
    resultDict = json.loads(result.content)

    for curResult in resultDict:
      changeNotification = ChangeNotificationMessage(dict=curResult, clientAPIInstance=self.clientAPIInstance)
      self.processMessage(changeNotification=changeNotification)
      if self.lastProcessedID is not None:
        self.lastProcessedID = changeNotification.messageID

    return remainingMessages

  def processMessage(self, changeNotification):
    pass


class EthosChangeNotificationPollerThreadQueueMode(EthosChangeNotificationPollerThread):
  pollerQueue = None

  def __init__(
    self,
    clientAPIInstance,
    loginSession,
    frequency,
    pageLimit,
    maxRequests,
    pollerQueue
  ):
    super().__init__(
      clientAPIInstance=clientAPIInstance,
      loginSession=loginSession,
      frequency=frequency,
      pageLimit=pageLimit,
      maxRequests=maxRequests,
      lastProcessedID=None
    )
    self.pollerQueue = pollerQueue

  def processMessage(self, changeNotification):
    self.pollerQueue.put(changeNotification)

class EthosChangeNotificationPollerThreadFunctionMode(EthosChangeNotificationPollerThread):
  messageProcessingFunction = None

  def __init__(
    self,
    clientAPIInstance,
    loginSession,
    frequency,
    pageLimit,
    maxRequests,
    lastProcessedID,
    messageProcessingFunction
  ):
    super().__init__(
      clientAPIInstance=clientAPIInstance,
      loginSession=loginSession,
      frequency=frequency,
      pageLimit=pageLimit,
      maxRequests=maxRequests,
      lastProcessedID=lastProcessedID
    )
    self.messageProcessingFunction = messageProcessingFunction

  def processMessage(self, changeNotification):
    self.messageProcessingFunction(apiClient=self.clientAPIInstance, messageid=changeNotification.messageID, changeNotification=changeNotification)



import threading
import time
import datetime
import pytz

class WorkerThreadExceptionClass(Exception):
  pass

#First impleimentaion was using pref_counter
# reimplemented to use date time due to limitations
def getNextWorkerTime(lastRunTime, curTime, frequency):
  retVal = lastRunTime
  while retVal < curTime:
    retVal = retVal + datetime.timedelta(seconds=int(frequency))
  return retVal


class WorkerThread(threading.Thread):
  sleepTime = None
  running = None
  stopped = None
  thrownException = None
  frequency = None

  nextWorkerTime = None

  def __init__(self, sleepTime, frequency):
    super().__init__()
    self.sleepTime = sleepTime
    self.running = False
    self.thrownException = None
    self.stopped = False
    self.frequency=frequency

    #First run is immediately when thread is started
    self.nextWorkerTime = datetime.datetime.now(pytz.timezone("UTC"))

  def run(self):
    try:
      self.stopped=False
      self.running = True
      while self.running:
        #now = datetime.datetime.now(pytz.timezone("UTC"))
        if datetime.datetime.now(pytz.timezone("UTC")) > self.nextWorkerTime:
          self.worker()
          self.nextWorkerTime = getNextWorkerTime(
            lastRunTime=self.nextWorkerTime,
            curTime=datetime.datetime.now(pytz.timezone("UTC")),
            frequency=self.frequency
          )
        time.sleep(self.sleepTime)
      self.stopped=True
    except Exception as Excep:
      self.thrownException = WorkerThreadExceptionClass("Poller thread threw an exception - " +  str(Excep))
      self.running = False

  def close(self, wait=True):
    if self.stopped:
      raise WorkerThreadExceptionClass("Should not try to close thread twice")
    self.stopped = False
    self.running = False
    if wait:
      while self.stopped == False:
        self.healthCheck()
        time.sleep(self.sleepTime)

  def healthCheck(self):
    #Run in controller thread to raise any exceptions
    if self.running:
      if not self.isAlive():
        raise WorkerThreadExceptionClass("Poller thread is no longer alive")
    if self.thrownException is not None:
      raise self.thrownException

  def worker(self):
    pass #shuold be overridden



from .ResPersons import registerKnownResourses as registerPersonsKnownResourses
from .BaseResourceWrapper import BaseResourceWrapper

knownResourses = {}
registerPersonsKnownResourses(knownResourses)

def getResourceWrapper(clientAPIInstance, dict, version, resourseName):
  if resourseName in knownResourses:
    if knownResourses[resourseName](version) is not None:
      return knownResourses[resourseName](version)(clientAPIInstance, dict, version, resourseName)
  return BaseResourceWrapper(clientAPIInstance, dict, version, resourseName)



from dateutil.parser import parse
import pytz
from .ResourceWrappers import getResourceWrapper

class ChangeNotificationMessage():
  clientAPIInstance = None
  messageID = None
  published = None
  operation = None
  resourceName = None
  resourceID = None
  resourceVersion = None
  resourceWrapper = None

  #dict = None

  def __init__(self, dict, clientAPIInstance):
    self.clientAPIInstance = clientAPIInstance
    self.messageID = dict["id"]

    dt = parse(dict["published"])
    self.published = dt.astimezone(pytz.utc)
    #print("rec:", dict["published"])
    #print("obj:", self.published)
    #print("out:", self.published.isoformat())

    self.operation = dict["operation"] #created, deleted, etc

    #These fields are always present - even for deleted operation
    self.resourceName = dict["resource"]["name"]
    self.resourceID = dict["resource"]["id"]
    if "version" in dict["resource"]:
      #version not sent for deleted operation
      self.resourceVersion = clientAPIInstance.getVersionIntFromHeader(dict["resource"]["version"])

    # for operation=deleted, the contentType is sent as empty
    # for operation=created, the content type is sent as resource-representation
    if dict["contentType"] == "resource-representation":
      self.resourceWrapper = getResourceWrapper(
        clientAPIInstance=self.clientAPIInstance,
        dict=dict["content"],
        version=self.resourceVersion,
        resourseName=self.resourceName
      )
    else:
      self.resourceWrapper=None

    #print("dict", dict)

    #Example message
    #{'id': '105',
    # 'published': '2020-06-23 09:26:24.732506+00',
    # 'resource': {
    #     'name': 'person-holds',
    #     'id': '7620b51d-acee-44cf-a4f9-85cc16f5c737',
    #     'version': 'application/vnd.hedtech.integration.v6+json'
    # },
    # 'operation': 'created',
    # 'contentType': 'resource-representation',
    # 'content': {
    #     'endOn': '2099-12-31T00:00:00Z',
    #     'id': '7620b51d-acee-44cf-a4f9-85cc16f5c737',
    #     'person': {'id': 'f1d31dd9-b141-4932-a8b6-d09637f1876d'},
    #     'startOn': '2020-01-17T00:00:00Z',
    #     'type': {
    #         'category': 'academic',
    #         'detail': {
    #             'id': '45182c89-6bb8-4996-a05a-6fce28f028eb'
    #         }
    #     }
    # },
    # 'publisher': {
    #     'id': '84592449-455b-40ea-9856-cb9e6878bd21',
    #     'applicationName': 'Imperial Student API - BILD'
    # }
    #}


from .APIClients import APIClientBase
from .EthosLoginSession import EthosLoginSessionBasedOnAPIKey
from .ResourceWrappers import getResourceWrapper
from .ResourceIterator import ResourceIterator
from .EthosChangeNotificationPollerThread import EthosChangeNotificationPollerThreadQueueMode, EthosChangeNotificationPollerThreadFunctionMode
import json

class CanNotStartChangeNotificationPollerTwiceException(Exception):
  pass

class EllucianEthosAPIClient(APIClientBase):
  refreshAuthTokenIfRequired = None
  changeNotificationPollerThread = None

  def __init__(self, baseURL, mock=None):
    super().__init__(baseURL=baseURL, mock=mock, forceOneRequestAtATime=True)
    self.changeNotificationPollerThread = None

  def getLoginSessionFromAPIKey(self, apiKey):
    return EthosLoginSessionBasedOnAPIKey(APIClient=self, apikey=apiKey)

  def _getResourceRAW(self, loginSession, resourceName, resourceID, version=None):
    #print("_getResourceRAW", resourceName, resourceID, version)

    def injectHeaderFN(headers):
      if version is not None:
        headers["Accept"] = "application/vnd.hedtech.integration.v" + version + "+json"

    result = self.sendGetRequest(
      url="/api/" + resourceName + "/" + resourceID,
      loginSession=loginSession,
      injectHeadersFn=injectHeaderFN
    )
    if result.status_code == 404:
      return (None, None, None)
    if result.status_code != 200:
      self.raiseResponseException(result)

    versionReturned = self.getVersionIntFromHeader(result.headers["x-hedtech-media-type"])
    return (result.content, versionReturned, resourceName)

  #Doc list https://xedocs.ellucian.com/xe-banner-api/ethos_apis/foundation/persons/person_get_guid_v6.html
  def getResource(self, loginSession, resourceName, resourceID, version=None):
    (resultContent, versionReturned, resourceName) = self._getResourceRAW(loginSession=loginSession, resourceName=resourceName, resourceID=resourceID, version=version)
    if resultContent is None:
      return None
    return getResourceWrapper(clientAPIInstance=self, dict=json.loads(resultContent), version=versionReturned, resourseName=resourceName)

  def getResourceIterator(self, loginSession, resourceName, version=None, pageSize=25, params=None):
    return ResourceIterator(self, loginSession, resourceName, version, pageSize, params=params)

  def getVersionIntFromHeader(self, meaidTypeHeaderValue):
    #example: application/vnd.hedtech.integration.v6+json
    requiredStart = "application/vnd.hedtech.integration.v"
    requiredEnd = "+json"
    if not meaidTypeHeaderValue.startswith(requiredStart):
      raise Exception("Could not determine resource version")
    meaidTypeHeaderValue = meaidTypeHeaderValue[len(requiredStart):]
    if not meaidTypeHeaderValue.endswith(requiredEnd):
      raise Exception("Could not determine resource version - header didn't end with " + requiredEnd)
    meaidTypeHeaderValue = meaidTypeHeaderValue[:-len(requiredEnd)]
    return meaidTypeHeaderValue

  def createResource(
    self,
    loginSession,
    resourceName,
    resourceDict,
    version=None
  ):
    if version is None:
      raise Exception("Must supply version when creating resource")
    def injectHeaderFN(headers):
        headers["Accept"] = "application/vnd.hedtech.integration.v" + version + "+json"
        headers["Content-Type"] = "application/vnd.hedtech.integration.v" + version + "+json"

    result = self.sendPostRequest(
      url="/api/" + resourceName,
      loginSession=loginSession,
      injectHeadersFn=injectHeaderFN,
      data=json.dumps(resourceDict)
    )
    if result.status_code != 201:
      self.raiseResponseException(result)

    versionReturned = self.getVersionIntFromHeader(result.headers["x-hedtech-media-type"])

    return getResourceWrapper(clientAPIInstance=self, dict=json.loads(result.content), version=versionReturned, resourseName=resourceName)

  def deleteResource(
    self,
    loginSession,
    resourceName,
    resourceID
  ):
    url = "/api/" + resourceName + "/" + resourceID

    result = self.sendDeleteRequest(
      url=url,
      loginSession=loginSession,
      injectHeadersFn=None
    )
    if result.status_code != 200:
      self.raiseResponseException(result)

  def startChangeNotificationPollerThread(
    self,
    loginSession,
    frequency, #number of seconds between fetches
    pageLimit, #number of change notifications to get per requests
    maxRequests, #maximum number of rquests to use in each fecth
    pollerQueue
  ):
    if self.changeNotificationPollerThread is not None:
      raise CanNotStartChangeNotificationPollerTwiceException()
    self.changeNotificationPollerThread = EthosChangeNotificationPollerThreadQueueMode(
      clientAPIInstance=self,
      loginSession=loginSession,
      frequency=frequency,
      pageLimit=pageLimit,
      maxRequests=maxRequests,
      pollerQueue=pollerQueue
    )
    self.changeNotificationPollerThread.start()

  def startChangeNotificationPollerThreadInFunctionMode(
    self,
    loginSession,
    frequency,  # number of seconds between fetches
    pageLimit,  # number of change notifications to get per requests
    maxRequests,  # maximum number of rquests to use in each fecth
    lastProcessedID,
    messageProcessingFunction
  ):
    if self.changeNotificationPollerThread is not None:
      raise CanNotStartChangeNotificationPollerTwiceException()
    self.changeNotificationPollerThread = EthosChangeNotificationPollerThreadFunctionMode(
      clientAPIInstance=self,
      loginSession=loginSession,
      frequency=frequency,
      pageLimit=pageLimit,
      maxRequests=maxRequests,
      lastProcessedID=lastProcessedID,
      messageProcessingFunction=messageProcessingFunction
    )
    self.changeNotificationPollerThread.start()

  def healthCheck(self):
    if self.changeNotificationPollerThread is not None:
      self.changeNotificationPollerThread.healthCheck()

  def close(self):
    if self.changeNotificationPollerThread is not None:
      self.changeNotificationPollerThread.close()
      self.changeNotificationPollerThread = None



# EllucianEthosPythonClient


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions

from .EllucianEthosAPIClient import EllucianEthosAPIClient, CanNotStartChangeNotificationPollerTwiceException

# imports to allow unit tests
from .WorkerThread import getNextWorkerTime




from builtins import list

from requests.models import Response
import base64
import json
import copy

# import python_Testing_Utilities
# In future I want a function like this in python_Testing_Utilities to give a good comparison of dicts
# def dictDifferenceString(
#           dictAName,
#           dictA,
#           dictBName,
#           dictB,
#           maxRecurse=3
#         ):
#   if maxRecurse==0:
#     return "Recurse limit reached\n"
#   aKeys = list(dictA.keys())
#   bKeys = list(dictB.keys())
#   if not python_Testing_Utilities.objectsEqual(aKeys, bKeys):
#     ret = "Keys mismatch\n"
#     ret += dictAName + ":" + str(aKeys) + "\n"
#     ret += dictBName + ":" + str(bKeys) + "\n"
#     return ret
#   # #list(dictA.keys()) == list(dictB.keys())
#   ret = ""
#   for curKey in aKeys:
#     try:
#       ret += "****" + str(dictB[curKey])  + "\n"
#       if not python_Testing_Utilities.objectsEqual(dictA[curKey],dictB[curKey]):
#         ret += dictAName + "." + curKey + " is not equal to " + dictBName + "." + curKey + "\n"
#     except Exception as err:
#       ret += str(err) + "\n"
#   return ret


class mockRequestClass():
  method = None
  url = None
  def __init__(self, method, url):
    self.method = method
    self.url = url

class ExpectedResult():
  reqFnName = None
  url = None
  data = None
  ignoreData = None
  def __init__(self, reqFnName, url, data, response=None, ignoreData=False):
    self.reqFnName = reqFnName
    self.url = url
    self.data = data
    self.response = response
    self.ignoreData = ignoreData

  def __str__(self):
    return  self.reqFnName + "->" + self.url

  def requestMatches(self, reqFnName, url, data):
    if self.reqFnName != reqFnName:
      return False
    if self.url != url:
      return False
    if self.ignoreData:
      return True
    if self.data != data:
      return False
    return True

  def getResponse(self):
    return self.response




class MockNoResultSetupException(Exception):
  recievedRequest = None
  def __init__(self, recievedRequest):
    self.recievedRequest = recievedRequest

  def __str__(self):
    return "NoMockResultsSetupException - " + str(self.recievedRequest)

class MockIncorrectResultException(Exception):
  recievedRequest = None
  expectedResult = None
  recievedData = None
  def __init__(self, recievedRequest, expectedResult, recievedData):
    self.recievedRequest = recievedRequest
    self.expectedResult = expectedResult
    self.recievedData = recievedData

  def isStrValidDICT(self, data):
    if not isinstance(data, str):
      return False
    try:
      x = json.loads(data)
      return True
    except Exception as exp:
      return False

  def __str__(self):
    if str(self.recievedRequest) == str(self.expectedResult):
      ret = "\n*******************************************\n"
      ret += "Request " + str(self.recievedRequest) + "\n"
      jsonComparison = False
      if self.isStrValidDICT(self.expectedResult.data):
        if self.isStrValidDICT(self.recievedData):
          jsonComparison = True
      ret += "Got Data:" + str(self.recievedData) + "\n\n"
      ret += "Expected Data:" + str(self.expectedResult.data) + "\n"
      if jsonComparison:
        ret += "Both results are valid json\n"
        # ret += dictDifferenceString(
        #   dictAName="GOT",
        #   dictA=json.loads(self.recievedData),
        #   dictBName="EXPECTED",
        #   dictB=json.loads(self.expectedResult.data)
        # )

      ret += "MockIncorrectResultException - Got call for " + str(self.recievedRequest) + " but data value mismatched"
      return ret
    return "MockIncorrectResultException - Got call for " + str(self.recievedRequest) + " but expected " + str(self.expectedResult)

class MockUnusedResultsException(Exception):
  unusedResults = None
  def __init__(self, unusedResults):
    self.unusedResults = unusedResults

  def __str__(self):
    return "There were " + str(len(self.unusedResults)) + " mock results not used"

class MockClass():
  stackOfExpectedReturnValues = None
  def __init__(self):
    self.stackOfExpectedReturnValues = []

  def registerNextResponse(
    self,
    reqFnName,
    url,
    data,
    #code,
    #error_type,
    status_code,
    contentBytes,
    contentHeaders=None,
    ignoreData=False
  ):
    the_response = Response()
    the_response.request = mockRequestClass(method=reqFnName, url=url)
    #the_response.code = code
    #the_response.error_type = error_type
    the_response.status_code = status_code
    if contentBytes is None:
      the_response._content = b""
    else:
      the_response._content = base64.b64decode(contentBytes)
    if contentHeaders != None:
      the_response.headers=contentHeaders
    self.stackOfExpectedReturnValues.append(
      ExpectedResult(
        reqFnName=reqFnName,
        url=url,
        data=data,
        response=the_response,
        ignoreData=ignoreData
      )
    )

  def returnNextResult(
    self,
    reqFnName,
    url,
    data
  ):
    recievedRequest = ExpectedResult(reqFnName, url, data)
    if (len(self.stackOfExpectedReturnValues)==0):
      raise MockNoResultSetupException(recievedRequest=recievedRequest)
    expectedResult = self.stackOfExpectedReturnValues.pop()
    if not expectedResult.requestMatches(reqFnName=reqFnName, url=url, data=data):
      raise MockIncorrectResultException(recievedRequest=recievedRequest, expectedResult=expectedResult, recievedData=data)
    return expectedResult.getResponse()

  def testComplete(self):
    if (len(self.stackOfExpectedReturnValues)!=0):
      print("Test complete but unused mock resutls - unwinding stack")
      origStack = copy.deepcopy(self.stackOfExpectedReturnValues)
      while (len(self.stackOfExpectedReturnValues) > 0):
        expectedResult = self.stackOfExpectedReturnValues.pop()
        print(expectedResult)

      raise MockUnusedResultsException(origStack)



import json
from .ResourceWrappers import getResourceWrapper

class ResourceIterator:
  apiClient = None
  loginSession = None
  resourceName = None
  version = None
  pageSize = None
  curList = None
  curIdx = None
  curOffset = None
  versionReturned = None
  params = None

  def __init__(self, apiClient, loginSession, resourceName, version, pageSize, params):
    self.apiClient = apiClient
    self.loginSession = loginSession
    self.resourceName = resourceName
    self.version = version
    self.pageSize = pageSize

    self.curList = []
    self.curIdx = 0
    self.curOffset = 0

    if params is None:
      self.params = {}
    else:
      self.params = params

    self.versionReturned = None

  def __iter__(self):
    return self

  def __next__(self):
    if self.curIdx >= len(self.curList):
      self.collectNextPage()
      if self.curIdx >= len(self.curList):
        raise StopIteration
    cur = self.curIdx
    self.curIdx += 1
    return getResourceWrapper(clientAPIInstance=self, dict=self.curList[cur], version=self.versionReturned,resourseName=self.resourceName)

  def collectNextPage(self):
    def injectHeaderFN(headers):
      if self.version is not None:
        headers["Accept"] = "application/vnd.hedtech.integration.v" + self.version + "+json"

    self.params["limit"] = str(self.pageSize)
    self.params["offset"] = str(self.curOffset)
    result = self.apiClient.sendGetRequest(
      url="/api/" + self.resourceName,
      params=self.params,
      loginSession=self.loginSession,
      injectHeadersFn=injectHeaderFN
    )
    if result.status_code != 200:
      self.apiClient.raiseResponseException(result)

    if self.versionReturned is None:
      self.versionReturned = self.apiClient.getVersionIntFromHeader(result.headers["x-hedtech-media-type"])

    self.curList = json.loads(result.content)
    self.curIdx = 0
    self.curOffset += len(self.curList)




from .BaseResourceWrapper import BaseResourceWrapper
import copy

def registerKnownResourses(resoursRegistryDict):
  def getKnownResource(version):
    majorVersion = version.split(".")[0]
    if majorVersion=="6":
      return PersonsV6
    if majorVersion=="8":
      return PersonsV8
    if majorVersion=="12":
      return PersonsV12
    return None
  resoursRegistryDict["persons"]=getKnownResource

class Persons(BaseResourceWrapper):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

  def _getDictForPut(self):
    retVal = copy.deepcopy(self.dict)

    #Updating addresses not yet supported
    # sending them to API causes error
    del retVal["addresses"]
    return retVal

class PersonsV6(Persons):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

class PersonsV8(Persons):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

class PersonsV12(Persons):
  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)



from setuptools import setup
import versioneer

#Dependancy lists maintained here and in tox.ini
sp_install_requires = [
#  'stomp.py==6.0.0'
]
sp_tests_require = [
  'nose==1.3.7',
  'python_Testing_Utilities==0.1.5'
]

all_require = sp_install_requires + sp_tests_require

setup(name='EllucianEthosPythonClient',
      version=versioneer.get_version(),
      cmdclass=versioneer.get_cmdclass(),
      description='Python package which provides Ellucian Ethos Client',
      url='https://github.com/rmetcalf9/EllucianEthosPythonClient',
      author='Robert Metcalf',
      author_email='rmetcalf9@googlemail.com',
      license='MIT',
      packages=['EllucianEthosPythonClient', 'EllucianEthosPythonClient.APIClients', 'EllucianEthosPythonClient.ResourceWrappers'],
      zip_safe=False,
      install_requires=sp_install_requires,
      tests_require=sp_tests_require,
      include_package_data=True)



import json
import copy

class BaseResourceWrapper():
  clientAPIInstance = None
  dict = None
  version = None
  resourceName = None
  resourceID = None
  def __init__(self, clientAPIInstance, dict, version, resourceName):
    self.clientAPIInstance = clientAPIInstance
    self.dict = dict
    self.version = version
    self.resourceName = resourceName

    if "id" not in dict:
      raise Exception("Invalid resource dict (missing id)")
    if not isinstance(version , str):
      raise Exception("Version passed must be a string")

    self.resourceID = dict["id"]

  def getMajorVersion(self):
    return self.version.split(".")[0]

  def _getDictForPut(self):
    return copy.deepcopy(self.dict)

  def save(self, loginSession):
    def injectHeaderFN(headers):
      headers["Accept"] = "application/vnd.hedtech.integration.v" + self.getMajorVersion() + "+json"
      headers["Content-Type"] = "application/vnd.hedtech.integration.v" + self.getMajorVersion() + "+json"
      ##print(headers)

    url = "/api/" + self.resourceName + "/" + self.resourceID

    result = self.clientAPIInstance.sendPutRequest(
      url=url,
      loginSession=loginSession,
      injectHeadersFn=injectHeaderFN,
      data=json.dumps(self._getDictForPut())
    )
    if result.status_code != 200:
      self.clientAPIInstance.raiseResponseException(result)

    self.dict = json.loads(result.content)

  def delete(self, loginSession):
    url = "/api/" + self.resourceName + "/" + self.resourceID

    result = self.clientAPIInstance.sendDeleteRequest(
      url=url,
      loginSession=loginSession,
      injectHeadersFn=None,
    )
    if result.status_code != 200:
      self.clientAPIInstance.raiseResponseException(result)

  def refresh(self, loginSession):
    (resultContent, versionReturned, resourceName) = self.clientAPIInstance._getResourceRAW(
      loginSession=loginSession,
      resourceName=self.resourceName,
      resourceID=self.resourceID,
      version=self.version
    )
    if versionReturned is None:
      raise Exception("Refresh failed - no resource found")
    if versionReturned != self.version:
      raise Exception("Internal Error wrong version on refresh")
    if resourceName != self.resourceName:
      raise Exception("Internal Error wrong resourceName on refresh")
    self.dict = json.loads(resultContent)
